#    Copyright (C) 2003-2007  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
__doc__ = """$id$"""
__version__='$Revision: 81 $'[11:-2]

import Globals, string, Products, inspect, pydoc, os, sys
from AccessControl import getSecurityManager, ClassSecurityInfo
from AccessControl.Permissions import change_configuration, view, view_management_screens, \
     manage_properties, change_permissions, undo_changes, access_contents_information
from AccessControl.unauthorized import Unauthorized
from Acquisition import aq_base
try:
    from Products.CMFPlone.PloneFolder import PloneFolder as Folder
except:
    try:
        from Products.OrderedFolder.OrderedFolder import OrderedFolder as Folder
    except:
        from OFS.Folder import Folder
    
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import zLOG
from utils import all_modules, pydoc_decode, locate
from ZpydocRenderer import RendererFolder
from Permissions import *

#
# Plone skins are optional ...
#
try:
    from Products.CMFCore.DynamicType import DynamicType
    from Products.CMFCore.utils import _getViewFor
    # need DynamicType for skins, and SimpleItem for context/path info etc
    class PortalContent ( DynamicType, SimpleItem ): pass
    DO_PLONE = 1
except:
    class PortalContent: pass
    DO_PLONE = 0

# do we have SilverCity (for sytax high-lighting)
try:
    from Products.PloneSilverCity.source import generate_html
    HAS_SILVERCITY = 1
except ImportError:
    HAS_SILVERCITY = 0

factory_type_information = (
    {
    'id'             : 'Zpydoc'
    , 'meta_type'      : 'Zpydoc'
    , 'description'    : """An automatic documentation tool."""
    , 'icon'           : 'zpydoc.gif'
    , 'product'        : 'Zpydoc'
    , 'factory'        : ''
    , 'implicitly_addable': 1
    , 'filter_content_types' : 1
    , 'allowed_content_types': ['ZpydocPackage']
    , 'immediate_view' : 'zpydoc_view'
    , 'global_allow'   : 0
    , 'actions'        : ( { 'id'            : 'view'
                             , 'name'        : 'View'
                             , 'action'      : 'string:${object_url}/zpydoc_view'
                             , 'permissions' : (view,)
                             , 'category'    : 'object'
                             }, )
    },
    {
    'id'             : 'ZpydocPage'
    , 'meta_type'      : 'ZpydocPage'
    , 'description'    : """An automatic documentation page."""
    , 'icon'           : 'package.png'
    , 'product'        : 'Zpydoc'
    , 'factory'        : ''
    , 'filter_content_types' : 1
    , 'allowed_content_types': []
    , 'immediate_view' : 'zpydocumentable_view'
    , 'global_allow'   : 0
    , 'actions'        : (
    { 'id'            : 'view'
      , 'name'        : 'View'
      , 'action'      : 'string:${object_url}/zpydocumentable_view'
      , 'permissions' : (view,)
      , 'category'    : 'object'
      },
    { 'id'            : 'src'
      , 'name'        : 'Source'
      , 'action'      : 'string:${object_url}/zpydocumentable_src'
      , 'condition'   : 'python: object.isfile()'
      , 'permissions' : (view_zpydoc_source,)
      , 'category'    : 'object'
      },
    )
    },
)

class ZpydocPage( PortalContent ):
    """
    a publishable unit (required for ZPublisher ...)

    note - this is not persisted !!
    """
    isPortalContent = 1
    meta_type = portal_type = 'ZpydocPage'

    _security = ClassSecurityInfo()
    _security.declareObjectPublic()

    __ac_permissions__ = PortalContent.__ac_permissions__ + (
        (view, ('getPackageInfo', 'isfile',),),
        (view_zpydoc_source, ('render_source',)),
        )

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, object, name,  renderer):
        #zLOG.LOG('ZpydocPage', zLOG.DEBUG, '__init__')
        self.id = name
        self.title = name.split('.')[-1]
        self.object = object
        self.name = name
        self.documentable = renderer.aq_parent
        self.renderer = renderer

    # force __call__ - we don't want aquisition of parents ...
    index_html = None

    def __call__(self, REQUEST, *args, **kw):
        """
        this will get invoked by ZPublisher to render the page
        """
        #zLOG.LOG('ZpydocPage', zLOG.DEBUG, '__call__()')
        #
        # do our 'file' permission security check (we have only implemented 1st level though ...)
        #
        user = getSecurityManager().getUser()
        difference = self.documentable.path_difference('.')
        if difference and difference != 'builtin__':
            pkg = string.replace(self.name, difference, '', 1).split('.')[0]
        else:
            pkg = self.name.split('.')[0]
        #zLOG.LOG('ZpydocPage', zLOG.DEBUG, '__call__() user=%s, package=%s' % (user.getUserName(), pkg))
        # the renderer has it's ZpyDocumentable as it's parent ...
        if self.documentable.isPermitted(pkg, user):
            if DO_PLONE:
                view = _getViewFor(self)
                result = apply(view, (self, self.REQUEST))
                return result
            # oh well, do old behaviour ...
            return self.renderer( self.object, self.name )
        else:
            #
            # force login 
            #
            raise Unauthorized()


    def isfile(self):
        """
        return true if is a file (otherwise we need security assertions on os ...)
        """
        return getattr(aq_base(self.object), '__file__', None)


    def render_source(self,plain=False):
        """
        render html source text
        """
        # file IO is private ;) - and we're going to be safe and calculate it, not
        # trusting our user to pass it in ...
        filename = inspect.getsourcefile(self.object)
        if plain:
            return open(filename,'r').read()
        if HAS_SILVERCITY:
            return generate_html(open(filename,'r').read(),'python')[0]
        return getattr(self, 'renderSource')(lines=open(filename, 'r').readlines())

    def getPackageInfo(self):
        """
        we need to wrap PackageInfo's in an acquisition context for permissions to work
        """
        # delegate to skin so end-user can additionally embellish result
        return self.createPackageInfo(self.object, self.name).__of__(self)

Globals.InitializeClass(ZpydocPage)


class Zpydoc(Folder):
    """
    Run's pydoc within the Zope HTTP server environment

    If you have the *OrderedFolder* product installed, you can control the order in
    which your documentables appear in the index_html

    If you want to customise the documentation, or want to do cool stuff with Zope
    documentation, then you're at the right place!
    """
    meta_type = portal_type = 'Zpydoc'
    id = 'Zpydoc'
    title = 'Zope PyDoc'
    icon = 'www/pydoc.gif'
    __replaceable__ = Globals.REPLACEABLE
    _security = ClassSecurityInfo()
    _reserved_names = ('index_html',)
    default_page = 'zpydoc_view'
    property_extensible_schema__ = 1
    __allow_access_to_unprotected_subobjects__ = 1
    
    __ac_permissions__ = (
        (view_management_screens, ('manage_main','manage_findFrame', 'manage_findForm',
                                   'manage_findAdv', 'manage_findResult', 'manage_renderers')),
        (view, ('', '__call__')),
        (manage_properties, ('manage_addProperty', 'manage_editProperties',
                             'manage_delProperties', 'manage_changeProperties',
                             'manage_propertiesForm', 'manage_propertyTypeForm',
                             'manage_changePropertyTypes', )),
        (access_contents_information,  ('hasProperty', 'propertyIds', 'propertyValues',
                                        'propertyItems', 'getProperty', 'getPropertyType',
                                        'propertyMap'), ('Anonymous', 'Manager') ),
        (change_permissions,  ('manage_access', 'permission_settings', 'ac_inherited_permissions',
                               'manage_roleForm', 'manage_role', 'manage_acquiredForm',
                               'manage_acquiredPermissions', 'manage_permissionForm',
                               'manage_permission', 'manage_changePermissions', 'permissionsOfRole',
                               'rolesOfPermission', 'acquiredRolesAreUsedBy', 'manage_defined_roles',
                               'userdefined_roles', 'manage_listLocalRoles', 'manage_editLocalRoles',
                               'manage_setLocalRoles', 'manage_addLocalRoles', 'manage_delLocalRoles' )),
        (undo_changes, ('manage_undo_transactions', 'undoable_transactions', 'manage_UndoForm' )),
        )
    
    manage_options = (
        {'label':'Contents', 'action':'manage_main', 'help':('Zpydoc', 'pydoc.stx') },
        {'label':'View', 'action':'' },
        {'label':'Renderers', 'action':'renderers/manage_workspace' },
        ) + Folder.manage_options[2:-2]

    index_html = PageTemplateFile('zpt/index_html', globals())
    
    def manage_renderers(self, REQUEST):
        """
        Redirect to renderers folder
        """
        REQUEST.RESPONSE.redirect('renderers/manage_main')

    def manage_delObjects(self, ids=[], REQUEST=None):
	"""
	this doesn't seem to properly route ...
	"""
	if ids:
	    Folder.manage_delObjects(ids)
	if REQUEST:
	    return self.manage_main(self, REQUEST)

    def __init__(self):
        self.renderers = RendererFolder('renderers')

    def all_meta_types(self):
        return filter(lambda x: x['name'] == 'ZpyDocumentable', Products.meta_types)

    _security.declareProtected(manage_zpydoc, 'all_documentables')
    def all_documentables(self):
        documentables = ['__builtin__']
        documentables.extend( all_modules() )
        return documentables

    _security.declareProtected(view, 'get_documentable')
    def get_documentable(self, obj, do_permissions=1):
	"""
	determine if this package is ZpyDocumentable and optionally that the user is 
        permissioned to	do so, returning the ZpyDocumentable (or None)
	"""
	if hasattr(obj, '__path__'):
            # it's a file system package ...
            path = getattr(obj, '__path__')
            if type(path) == type([]):
                path = path[0]
                directory,file = os.path.split( path )
        else:
            try:
                # it's a module ...
                directory,file = os.path.split( inspect.getsourcefile(obj) )
            except TypeError:
                # ok - it's a builtin (we could 100% verify this but ...)
                directory = 'builtin__'
            except:
                raise

        # step thru directory looking for a ZpyDocumentable to bind to ...
        zpydoc = None
        while not zpydoc and directory and directory != '/':
            try:
                zpydoc = filter( lambda x, y=directory: x.directory == y, self.objectValues() )[0]
            except:
                directory,file = os.path.split( directory )
            for documentable in self.objectValues('ZpyDocumentable'):
                if documentable.path() == directory:
                    zpydoc = documentable
                    break
        
	if zpydoc:
            difference = zpydoc.path_difference('.')
            if difference and difference != 'builtin__':
                pkg = string.replace(obj.__name__, difference, '', 1).split('.')[0]
            else:
                pkg = obj.__name__.split('.')[0]
            if do_permissions and not zpydoc.isPermitted(pkg, getSecurityManager().getUser()):
	        zpydoc = None

	return zpydoc

    _security.declareProtected(view, 'render')
    def render(self):
        #
        # driver for index_html
        #
        modules = []
        seen = {}
        for documentable in self.objectValues():
            if documentable.directory == 'builtin__':
                modules.append( documentable.renderer().builtins() )
            else:
                # TODO - no parameters into index() ...
                modules.append( documentable.renderer().index(documentable.directory, seen) )
        return string.join(modules, '')

    #def __bobo_traverse__(self, REQUEST, name):
    def __getitem__(self, name):
        """
        this is where all the magic resides - it's a factory binding mechanism:
         * parsing the 'url'
         * binding to our ZpyDocumentable directory for rendering and access permissions
         * assembling a ZpydocPage for authorising and publishing
        """
        name = pydoc_decode(name)
        if name[-5:] == '.html': name = name[:-5]
        if name[:1] == '/': name = name[1:]

        #zLOG.LOG('Zpydoc', zLOG.DEBUG, '__getitem__(%s)' % name)
        if name and name != '.':
            try:
                obj = locate(name, forceload=0)
            except pydoc.ErrorDuringImport, value:
                #return html_quote( str(value) )
                #zLOG.LOG('Zpydoc', zLOG.DEBUG, 'pyDoc.ErrorDuringImport(%s)' % value)
                return None
            if obj:
		zpydoc = self.get_documentable(obj, 0)
                if zpydoc:
		    #zLOG.LOG('Zpydoc', zLOG.DEBUG, 'rendering %s with %s' % (name, zpydoc.renderer()))
                    return ZpydocPage( obj, name , zpydoc.renderer() ).__of__(self)
                else:
                    #zLOG.LOG('Zpydoc', zLOG.DEBUG, 'no renderer %s [%s] ' % (obj, name))
                    raise AttributeError, 'no renderer for %s [%s]' % (obj, name)
                 
            else:
                #zLOG.LOG('Zpydoc', zLOG.DEBUG, 'no object returned from locate (%s)!\npath=%s ' % (name, sys.path))
                return Folder.__getitem__(self, name)
                
        return None

    _security.declareProtected(change_configuration, 'manage_repair')
    def manage_repair(self, REQUEST=None):
        """
        if we change the underlying structure of Zpydoc, then you can apply this method
        to repair an instance - this is designed to be rerunnable ...
        """
        if not hasattr(aq_base(self), 'renderers'):
            self.renderers = RendererFolder('renderers', '')

        for ob in self.objectValues('ZpyDocumentable'):
            ob._repair()

        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Object Repaired')
            return self.manage_main(self, REQUEST)

Globals.InitializeClass(Zpydoc)

def manage_addZpydoc(self, REQUEST=None):
    """
    Add a Zpydoc autogenerating Python/Zope documentation tool
    """
    self._setObject('Zpydoc', Zpydoc())
    if REQUEST:
        REQUEST.RESPONSE.redirect('%s/Zpydoc/manage_main' % REQUEST['URL3'])

