#    Copyright (C) 2003-2007  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Globals, zLOG, Products, inspect, os, sys, traceback
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view
from pydoc import ispackage
import PythonInfo
from PythonInfo import structured_text, getmembers
from Products.Zpydoc.utils import implementsMethod
from Products.Zpydoc.Permissions import view_zpydoc_source
from Acquisition import aq_base
from OFS.SimpleItem import SimpleItem
import OFS.misc_

# let it recognized PythonMethods as methods
inspect.ismethod= implementsMethod

class FunctionInfo( PythonInfo.FunctionInfo ):
    """
    We want to get Zope Permission info into the function info
    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()

    def __init__(self, function, permission):
        #zLOG.LOG('ZopeFunctionInfo', zLOG.DEBUG, permission)
        PythonInfo.FunctionInfo.__init__(self, function)
        self.permission = permission

    def arguments(self, formatfn=lambda x: '<span style="color:grey; font-size:10pt">=%s</span>' % repr(x)):
        """
        return an html-formatted list of the function arguments
        """
        try:
            return FunctionInfo.inheritedAttribute('args')(self, formatfn)
        except:
            # it's probably an interface method
            return '(...)'

Globals.InitializeClass(FunctionInfo)


class ClassInfo( PythonInfo.ClassInfo ):
    """
    We want to place additional functions to access 'Public' and 'Private'
    methods etc
    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()

    __ac_permissions__ = PythonInfo.ClassInfo.__ac_permissions__ + (
        (view, ('public_methods', 'private_methods')),
        )

    def __init__(self, cls):
        #zLOG.LOG('ZopeClassInfo', zLOG.DEBUG, '__init__(%s)' % (cls))
        self._class = cls
        self.id = cls.__name__
        self.title = cls.__name__
        #
        # Push Product Registry info into this namespace - well the 'productinfo'
        # namespace anyway.  We want to try and avoid potential name clashes if
        # possible ...
        #
        if hasattr(cls, 'meta_type'):
            self.productinfo = ProductInfo(cls.meta_type)
        else:
            self.productinfo = None

    def methods(self):
        # let's do some caching ...
        #if not hasattr(self, '_methods'):
        # find and assign the permission
        #permissions = getPermissionMapping()
        pm_syms = {}
        if hasattr(self._class, '__ac_permissions__'):
            for perms in getattr(self._class, '__ac_permissions__'):
                for fn in perms[1]:
                    pm_syms[fn] = perms[0]

        #self._methods = map ( lambda x,y=pm_syms: FunctionInfo(x[1].im_func, y.get(x[0], '')),
        #                      getmembers(self._class, inspect.ismethod) )
        self._methods = []
        try:
            methods = getmembers(self._class, inspect.ismethod) 
        except Exception, e:
            methods = []
        for name,method in methods:
            self._methods.append( FunctionInfo( method.im_func, pm_syms.get(name, '') ).__of__(self) )

        if hasattr(self._class, '_Interface__attrs'):
            self._methods.extend(map(lambda x,y=pm_syms,self=self: FunctionInfo(x[1], y.get(x[0], '')).__of__(self),
                                     self._class._Interface__attrs.items()))
        #zLOG.LOG('ZopeClassInfo', zLOG.DEBUG, 'methods: %s' % self._methods)
        return self._methods

    def public_methods(self):
        """
        methods not beginning with an underscore, and having a docstring and a permission
        assignment (otherwise requiring Manager role to access - which effectively makes
        it private)
        """
        return filter(lambda x: x.id[0] != '_' and hasattr(x, '__doc__') and x.permission,
                      self.methods())

    def private_methods(self):
        """
        methods beginning with an underscore, or without a docstring (or requiring Manager role)
        """
        return filter(lambda x: x.id[0] == '_' or not hasattr(x, '__doc__') or not x.permission,
                      self.methods())

Globals.InitializeClass(ClassInfo)
        
class ModuleInfo( PythonInfo.ModuleInfo ):
    """
    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()

    def __init__(self, module):
        #zLOG.LOG('ZopeModuleInfo', zLOG.DEBUG, '__init__(%s)' % (module))
        self.id = module.__name__
        self.title = module.__name__
        self.meta_type = getattr(module, 'meta_type', 'ModuleInfo')
        self._module = module
    
    def classes(self):
        #zLOG.LOG('ZopeModuleInfo', zLOG.DEBUG, 'classes')
        results = []
        for k,v in getmembers(self._object, inspect.isclass):
            if (inspect.getmodule(v) or self._object) is self._object:
                results.append((k, v))
        return map(lambda x,self=self: ClassInfo( x[1] ).__of__(self), results)

Globals.InitializeClass(ModuleInfo)
    
class PackageInfo( PythonInfo.PackageInfo ):
    """
    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, object, name):
        """
        This is the same as it's parent, but the scope of the Info's is different ...
        """
        #zLOG.LOG('ZopePackageInfo', zLOG.DEBUG, '__init__(%s) -->%s' % (name, object))
        self.id = name
        self.title = name
        self._object = object
        self._modpkgs = []

        if hasattr(object, '__path__'):
            #
            # it's a package - get modules from file system ...
            #
            seen = {}
            self.directory = object.__path__[0]
            #zLOG.LOG('PackageInfo', zLOG.DEBUG, 'directory = %s' % self.directory)
            files = os.listdir(self.directory)

            def found(name, ispackage, modpkgs=self._modpkgs, seen=seen):
                if not seen.has_key(name):
                    modpkgs.append({'name':name, 'ispackage':ispackage})
                    seen[name] = 1

            # Package spam/__init__.py takes precedence over module spam.py.
            for file in files:
                path = os.path.join(self.directory, file)
                if ispackage(path): found(file, 1)
            for file in files:
                path = os.path.join(self.directory, file)
                if os.path.isfile(path):
                    modname = inspect.getmodulename(file)
                    if modname: found(modname, 0)
        else:
            #
            # it's a module ...
            #
            #zLOG.LOG('PackageInfo', zLOG.DEBUG, 'doing module processing (%s)' % object)
            self._modpkgs = map( lambda x: ModuleInfo( x[1] ),
                                 getmembers(object, inspect.ismodule) )
        #zLOG.LOG('PackageInfo', zLOG.DEBUG, 'modules=%s' % self._modpkgs)
        self._modpkgs.sort()

    def classes(self):
        """
        return a list of class meta data objects in this module/package
        """
        #zLOG.LOG('ZopePackageInfo', zLOG.DEBUG, 'classes()')
        if inspect.isclass(self._object):
            return [ ClassInfo(self._object).__of__(self) ]
        results = []
        for k,v in getmembers(self._object, inspect.isclass):
            #zLOG.LOG('ZopePackageInfo', zLOG.DEBUG, 'classes(%s)' % v)
            if inspect.getmodule(v) == self._object:
                results.append(v)
        return map(lambda x,self=self: ClassInfo(x).__of__(self), results)


    def __getitem__(self, name):
        """
        check for python builtin's before 'normal' zope objects ...
        """
        return  PythonInfo.PackageInfo.__getitem__(self, name) or getattr(aq_base(self._object), name, '')

Globals.InitializeClass(PackageInfo)


class ProductInfo(SimpleItem):
    """
    product={'visibility': 'Global',
             'interfaces': [<Interface webdav.EtagSupport.EtagBaseInterface at 83fa108>],
             'container_filter': None,
             'action': 'manage_addProduct/Zpydoc/manage_addZpydoc',
             'permission': 'Manage Zpydocs',
             'name': 'Zpydoc',
             'product': 'Zpydoc',
             'instance': <extension class Products.Zpydoc.Zpydoc.Zpydoc at 922f208>}

    instance={'icon': 'misc_/Zpydoc/renderer.gif',
            'id': 'PyDoc', '_renderer': <PyHTMLDoc at 0x965a288>,
            '__call__': <function __call__ at 0x965a0e4>,
            '__implements__': (<Interface Products.Zpydoc.ZpydocRenderer.IZpydocRenderer at 92576c0>,),
            'meta_type': 'pydocRenderer',
            'directory': <function directory at 0x9659c14>,
            '__getattr__': <function __getattr__ at 0x9659c4c>,
            'title': 'Vanilla pydoc',
            'builtins': <function builtins at 0x965a164>,
            '__doc__': '\n    Render a standard pydoc page by delegating to pydoc\n\n    We have to be very careful about not explicitly defining methods here because\n    then they do not become overrideable via ObjectManager objects in derivations.\n    This is why we have used containment of the pydoc object.\n    ',
            '__module__': 'Products.Zpydoc.renderers.PyDoc'}

    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()
    __allow_access_to_unprotected_subobjects__ = 1

    __ac_permissions__ = SimpleItem.__ac_permissions__ + (
        (view, ('iconpath', 'icon', 'interfaces', 'doc',)),
        )

    def __init__(self, meta_type):
        #zLOG.LOG('ZopeProductInfo', zLOG.DEBUG, '__init__(%s)' % meta_type)
        self.id = meta_type
        self._product = None
        #
        # unfortunately, cls has a different address to the instance in Products.meta_types :(
        # and there's something naf where full module path resolution doesn't exist ...
        #
        # we should cache this maybe ???
        #
        for p in Products.meta_types:
            product =  p.get('instance', None)
            if product and hasattr(product, 'meta_type') and product.meta_type == meta_type:
                self._product = p
                break

        if self._product and self._product.has_key('interfaces'):
            interfaces = []
            for interface in self._product['interfaces']:
                    interfaces.append( InterfaceInfo(interface) )
            # cunninly mask the hash table with an attribute ...
            self._interfaces = interfaces
        else:
            self._interfaces = []
    
        #zLOG.LOG('ZopeProductInfo', zLOG.DEBUG, self._product)

    def iconpath(self):
        try:
            return self._product['instance'].icon
        except:
            return None
        
    def icon(self):
        #
        # hmmm - think this is brokrn ...
        #
        try:
            name=os.path.split(self.iconpath())[1]
            return getattr(OFS.misc_.misc_, self._product['product'])[name]
        except:
            return None

    def interfaces(self):
        """
        return list of interface classes
        """
        return map(lambda x,y=self: x.__of__(y), self._interfaces)
    
    def doc(self):
        """
        return the document string
        """
        msg = getattr(self._product, '__doc__' )
        if msg:
            return structured_text( msg )
        return ''
        
    def __getitem__(self, name):
        #zLOG.LOG('ZopeProductInfo', zLOG.DEBUG, '__getitem__(%s)' % name)
        item = self._product.get(name, None)
        return item

Globals.InitializeClass(ProductInfo)
       
class InterfaceInfo(ClassInfo):
    """
    a Zope Interface object
    '_Interface__attrs': {'http__etag': <Method instance at 8379c78>, 'http__refreshEtag': <Method instance at 8450790>},
    '__doc__': '    Basic Etag support interface, meaning the object supports generating\n
                    an Etag that can be used by certain HTTP and WebDAV Requests.\n    ',
    '__bases__': (<Interface Interface._Interface.Interface at 82d5b70>,),
    '__module__': 'webdav.EtagSupport',
    '__name__': 'EtagBaseInterface'
    """
    _security = ClassSecurityInfo()
    _security.declareObjectPublic()
    __allow_access_to_unprotected_subobjects__ = 1

    __ac_permissions__ = ClassInfo.__ac_permissions__ + (
        (view, ('methods', 'name',)),
        (view_zpydoc_source, ('getsourcefile',)),
        )

    def name(self):
        return self['name']

    def getsourcefile(self):
        return self['module']
    
    def methods(self):
        return map ( lambda x,self=self: FunctionInfo(x[1], '').__of__(self),
                     self._class._Interface__attrs.items() )

Globals.InitializeClass(InterfaceInfo)
